# frozen_string_literal: true

require 'httparty'
require 'openssl'
require 'time'
require 'digest/sha2'
require 'uri'

class DistributedPress
  module V1
    module Social
      # Social Distributed Press APIv1 client
      #
      # Inspired by Mastodon's Request
      #
      # @todo It'd be nice to implement this on HTTParty itself
      # @see {https://github.com/mastodon/mastodon/blob/main/app/lib/request.rb}
      class Client
        include ::HTTParty

        # Signing algorithm
        #
        # @todo is it possible to use other algorithms?
        ALGORITHM = 'rsa-sha256'
        # Required by HTTP Signatures
        REQUEST_TARGET = '(request-target)'
        # Headers included in the signature
        SIGNABLE_HEADERS = [REQUEST_TARGET, 'Host', 'Date', 'Digest']

        # API URL
        # @return [String]
        attr_reader :url

        # RSA key size
        #
        # @return [Integer]
        attr_reader :key_size

        # Public key URL
        #
        # @return [String]
        attr_reader :public_key_url

        # @param :url [String] Social Distributed Press URL
        # @param :public_key [String] URL where public key is available
        # @param :private_key_pem [String] RSA Private key, PEM encoded
        # @param :key_size [Integer]
        # @param :logger [Logger]
        def initialize(url: 'https://social.distributed.press', public_key_url:, private_key_pem: nil, key_size: 2048, logger: nil)
          self.class.default_options[:base_uri] = @url = HTTParty.normalize_base_uri(url)
          self.class.default_options[:logger] = logger
          self.class.default_options[:log_level] = :debug

          @public_key_url = public_key_url
          @key_size = key_size
          @private_key_pem = private_key_pem
        end

        # POST request
        #
        # @todo Use DRY-schemas
        # @param endpoint [String]
        # @param body [Hash]
        # @return [Hash]
        def post(endpoint:, body:)
          body = body.to_json
          headers = default_headers

          add_request_specific_headers! headers, 'post', endpoint
          checksum_body! body, headers
          sign_headers! headers

          self.class.post(endpoint, body: body, headers: headers)
        end

        # Loads or generates a private key
        #
        # @return [OpenSSL::PKey::RSA]
        def private_key
          @private_key ||= OpenSSL::PKey::RSA.new(@private_key_pem || key_size)
        end

        # Public key
        #
        # @return [OpenSSL::PKey::RSA]
        def public_key
          private_key.public_key
        end

        # Host
        #
        # @return [String]
        def host
          @host ||= URI.parse(url).host
        end

        private

        def default_headers
          {
            'User-Agent' => "DistributedPress/#{DistributedPress::VERSION}",
            'Content-Type' => 'application/ld+json; profile="https://www.w3.org/ns/activitystreams"',
            'Host' => host
          }
        end

        # HTTP Signatures
        #
        # @see {https://docs.joinmastodon.org/spec/security/}
        # @param :headers [Hash]
        def sign_headers!(headers)
          headers['Signature'] = {
            'keyId' => public_key_url,
            'algorithm' => ALGORITHM,
            'headers' => signable_headers(headers),
            'signature' => signed_headers(headers)
          }.map do |key, value|
            "#{key}=\"#{value}\""
          end.join(',')

          headers.delete REQUEST_TARGET

          nil
        end

        # List of headers to be signed, removing headers that don't
        # exist on the request.
        #
        # @return [String]
        def signable_headers(headers)
          (SIGNABLE_HEADERS & headers.keys).join(' ').downcase
        end

        # Sign headers
        #
        # @param :headers [Hash]
        # @return [String]
        def signed_headers(headers)
          Base64.strict_encode64(
            private_key.sign(
              OpenSSL::Digest.new('SHA256'),
              signature_content(headers)
            )
          )
        end

        # Generates a string to be signed
        #
        # @param :headers [Hash]
        # @return [String]
        def signature_content(headers)
          headers.slice(*SIGNABLE_HEADERS).map do |key, value|
            "#{key.downcase}: #{value}"
          end.join("\n")
        end

        # Generate a checksum for the request body
        #
        # @param :body [String]
        # @param :headers [Hash]
        def checksum_body!(body, headers)
          headers['Digest'] = "SHA-256=#{Digest::SHA256.base64digest body}"

          nil
        end

        # Headers specific to a single request
        #
        # @param :headers [Hash] Headers
        # @param :verb [String] HTTP verb
        # @param :endpoint [String] Path
        def add_request_specific_headers!(headers, verb, endpoint)
          headers['Date'] = Time.now.utc.httpdate
          headers[REQUEST_TARGET] = "#{verb} #{endpoint}"

          nil
        end
      end
    end
  end
end
